import copy
import uuid

import docker
import pytest

import substra
from substra.sdk import exceptions
from substra.sdk import models

from ...fl_interface import FLTaskInputGenerator
from ...fl_interface import FLTaskOutputGenerator
from ...fl_interface import FunctionCategory
from ...fl_interface import InputIdentifiers


def test_wrong_debug_spawner():
    with pytest.raises(exceptions.SDKException) as err:
        substra.Client(backend_type="test")
    assert (
        str(err.value) == "Unknown value for the execution mode: test, valid values are: dict_values"
        "([<BackendType.REMOTE: 'remote'>, <BackendType.LOCAL_DOCKER: 'docker'>, <BackendType.LOCAL_SUBPROCESS: "
        "'subprocess'>])"
    )


def test_get_backend_type_docker(docker_clients):
    assert docker_clients[0].backend_mode == substra.BackendType.LOCAL_DOCKER


def test_get_backend_type_subprocess(subprocess_clients):
    assert subprocess_clients[0].backend_mode == substra.BackendType.LOCAL_SUBPROCESS


def test_get_backend_type_remote():
    client = substra.Client(url="foo.com", backend_type="remote", token="foo")
    assert client.backend_mode == substra.BackendType.REMOTE


class TestsDebug:
    def test_client_tmp_dir(self, clients):
        """Test the creation of a temp directory for the debug client"""
        assert clients[0].temp_directory

    @pytest.mark.parametrize("dockerfile_type", ("BAD_ENTRYPOINT", "NO_ENTRYPOINT", "NO_FUNCTION_NAME"))
    def test_client_bad_dockerfile(self, asset_factory, dockerfile_type, clients):
        client = clients[0]

        dataset_query = asset_factory.create_dataset()
        dataset_1_key = client.add_dataset(dataset_query)

        data_sample = asset_factory.create_data_sample(datasets=[dataset_1_key])
        sample_1_key = client.add_data_sample(data_sample)

        function_query = asset_factory.create_function(FunctionCategory.simple, dockerfile_type=dockerfile_type)
        function_key = client.add_function(function_query)

        cp = asset_factory.create_compute_plan()
        cp.tasks = [
            substra.sdk.schemas.ComputePlanTaskSpec(
                function_key=function_key,
                data_manager_key=dataset_1_key,
                task_id=str(uuid.uuid4()),
                train_data_sample_keys=[sample_1_key],
                inputs=FLTaskInputGenerator.task(opener_key=dataset_1_key, data_sample_keys=[sample_1_key]),
                outputs=FLTaskOutputGenerator.traintask(),
                worker=client.organization_info().organization_id,
            ),
        ]
        with pytest.raises((substra.sdk.backends.local.compute.spawner.base.ExecutionError, docker.errors.APIError)):
            client.add_compute_plan(cp)

    def test_compute_plan_add_update_tasks(self, asset_factory, clients):
        client = clients[0]
        client = substra.Client(backend_type=substra.BackendType.LOCAL_SUBPROCESS)
        compute_plan = client.add_compute_plan(
            substra.sdk.schemas.ComputePlanSpec(
                key=str(uuid.uuid4()),
                tag=None,
                name="My ultra cool compute plan",
                metadata=dict(),
            )
        )

        assert compute_plan.status == models.ComputePlanStatus.empty

        dataset_query = asset_factory.create_dataset()
        dataset_key = client.add_dataset(dataset_query)

        data_sample = asset_factory.create_data_sample(datasets=[dataset_key])
        data_sample_key = client.add_data_sample(data_sample)

        function_query = asset_factory.create_function(FunctionCategory.simple)
        function_key = client.add_function(function_query)

        traintask_key = client.add_task(
            substra.sdk.schemas.TaskSpec(
                function_key=function_key,
                compute_plan_key=compute_plan.key,
                rank=None,
                metadata=None,
                inputs=FLTaskInputGenerator.task(opener_key=dataset_key, data_sample_keys=[data_sample_key]),
                outputs=FLTaskOutputGenerator.traintask(),
                worker=client.organization_info().organization_id,
            )
        )

        traintask = substra.sdk.schemas.ComputePlanTaskSpec(
            function_key=function_key,
            task_id=str(uuid.uuid4()),
            inputs=FLTaskInputGenerator.task(opener_key=dataset_key, data_sample_keys=[data_sample_key])
            + FLTaskInputGenerator.trains_to_train([traintask_key]),
            outputs=FLTaskOutputGenerator.traintask(),
            worker=client.organization_info().organization_id,
        )

        compute_plan = client.add_compute_plan_tasks(
            key=compute_plan.key,
            tasks={
                "traintasks": [traintask],
            },
        )

    def test_compute_plan_update(self, clients):
        client = clients[0]
        old_name = "My ultra cool compute plan"
        new_name = "My even cooler compute plan"

        compute_plan = client.add_compute_plan(
            substra.sdk.schemas.ComputePlanSpec(
                key=str(uuid.uuid4()),
                tag="cool-tag",
                name=old_name,
                metadata=dict(),
            )
        )

        client.update_compute_plan(
            key=compute_plan.key,
            name=new_name,
        )

        compute_plan.name = new_name
        updated_compute_plan = client.get_compute_plan(compute_plan.key)

        assert compute_plan == updated_compute_plan

    def test_tasks_extra_fields(self, asset_factory, clients):
        client = clients[0]

        # set dataset, metric and function
        dataset_query = asset_factory.create_dataset()
        dataset_key = client.add_dataset(dataset_query)

        data_sample = asset_factory.create_data_sample(datasets=[dataset_key])
        data_sample_key = client.add_data_sample(data_sample)

        function_query = asset_factory.create_function(FunctionCategory.simple)
        function_key = client.add_function(function_query)

        predict_function_spec = asset_factory.create_function(category=FunctionCategory.predict)
        predict_function_key = client.add_function(predict_function_spec)

        composite_function_query = asset_factory.create_function(FunctionCategory.composite)
        composite_function_key = client.add_function(composite_function_query)

        metric_query = asset_factory.create_function(category=FunctionCategory.metric)
        metric_key = client.add_function(metric_query)

        # test traintask extra field

        traintask_key = client.add_task(
            substra.sdk.schemas.TaskSpec(
                function_key=function_key,
                inputs=FLTaskInputGenerator.task(opener_key=dataset_key, data_sample_keys=[data_sample_key]),
                outputs=FLTaskOutputGenerator.traintask(),
                worker=client.organization_info().organization_id,
            )
        )

        traintask = client.get_task(traintask_key)
        dataset_ref = [x for x in traintask.inputs if x.identifier == substra.sdk.schemas.StaticInputIdentifier.opener]
        assert len(dataset_ref) == 1
        assert dataset_ref[0].asset_key == dataset_key
        assert len(traintask.inputs) == 2  # data sample + opener

        # test predicttask extra fields
        predicttask_key = client.add_task(
            substra.sdk.schemas.TaskSpec(
                function_key=predict_function_key,
                inputs=FLTaskInputGenerator.task(opener_key=dataset_key, data_sample_keys=[data_sample_key])
                + FLTaskInputGenerator.train_to_predict(traintask_key),
                outputs=FLTaskOutputGenerator.predicttask(),
                worker=client.organization_info().organization_id,
            )
        )

        predicttask = client.get_task(predicttask_key)
        dataset_ref = [
            x for x in predicttask.inputs if x.identifier == substra.sdk.schemas.StaticInputIdentifier.opener
        ]
        assert len(dataset_ref) == 1
        assert dataset_ref[0].asset_key == dataset_key

        assert len(predicttask.inputs) == 3  # data sample + opener + input task

        model_ref = [x for x in predicttask.inputs if x.identifier == InputIdentifiers.shared]
        assert len(model_ref) == 1
        assert model_ref[0].parent_task_key == traintask_key

        # test testtask extra fields
        testtask_key = client.add_task(
            substra.sdk.schemas.TaskSpec(
                function_key=metric_key,
                inputs=FLTaskInputGenerator.task(opener_key=dataset_key, data_sample_keys=[data_sample_key])
                + FLTaskInputGenerator.predict_to_test(predicttask_key),
                outputs=FLTaskOutputGenerator.testtask(),
                worker=client.organization_info().organization_id,
            )
        )
        testtask = client.get_task(testtask_key)

        dataset_ref = [
            x for x in predicttask.inputs if x.identifier == substra.sdk.schemas.StaticInputIdentifier.opener
        ]
        assert len(dataset_ref) == 1
        assert dataset_ref[0].asset_key == dataset_key
        assert len(predicttask.inputs) == 3  # data sample + opener + input task

        model_ref = [x for x in testtask.inputs if x.identifier == InputIdentifiers.predictions]
        assert len(model_ref) == 1
        assert model_ref[0].parent_task_key == predicttask_key

        assert testtask.function
        assert testtask.function.key == metric_key

        # test composite extra field

        composite_traintask_key = client.add_task(
            substra.sdk.schemas.TaskSpec(
                function_key=composite_function_key,
                inputs=FLTaskInputGenerator.task(opener_key=dataset_key, data_sample_keys=[data_sample_key]),
                outputs=FLTaskOutputGenerator.composite_traintask(),
                worker=client.organization_info().organization_id,
            )
        )

        composite_traintask = client.get_task(composite_traintask_key)
        dataset_ref = [
            x for x in composite_traintask.inputs if x.identifier == substra.sdk.schemas.StaticInputIdentifier.opener
        ]
        assert len(dataset_ref) == 1
        assert dataset_ref[0].asset_key == dataset_key
        assert len(composite_traintask.inputs) == 2  # data sample + opener

    def test_add_two_compute_plan_with_same_cp_key(self, clients):
        client = clients[0]
        common_cp_key = str(uuid.uuid4())
        client.add_compute_plan(
            substra.sdk.schemas.ComputePlanSpec(
                key=common_cp_key,
                tag=None,
                name="My compute plan",
                metadata=dict(),
            )
        )
        with pytest.raises(exceptions.KeyAlreadyExistsError):
            client.add_compute_plan(
                substra.sdk.schemas.ComputePlanSpec(
                    key=common_cp_key,
                    tag=None,
                    name="My other compute plan",
                    metadata=dict(),
                )
            )

    def test_add_compute_plan_with_wrong_metadata(self, clients):
        client = clients[0]
        cp_key = str(uuid.uuid4())
        with pytest.raises(exceptions.InvalidRequest):
            client.add_compute_plan(
                substra.sdk.schemas.ComputePlanSpec(
                    key=cp_key,
                    tag=None,
                    name="My compute plan",
                    metadata={"wrong__key": "value"},
                )
            )

    def test_permissions_consistency_add_dataset(self, clients, asset_factory):
        client = clients[0]

        permissions = substra.sdk.schemas.Permissions(public=False, authorized_ids=["foo"])
        copy_permissions = copy.deepcopy(permissions)

        dataset_query = asset_factory.create_dataset(permissions=permissions)

        client.add_dataset(dataset_query)

        assert str(copy_permissions) == str(permissions)

    def test_permissions_consistency_add_function(self, clients, asset_factory):
        client = clients[0]

        permissions = substra.sdk.schemas.Permissions(public=False, authorized_ids=["foo"])
        copy_permissions = copy.deepcopy(permissions)

        function_query = asset_factory.create_function(permissions=permissions)

        client.add_function(function_query)

        assert str(copy_permissions) == str(permissions)

    def test_live_performances_json_file_exist(self, asset_factory, clients):
        """Assert the performances file is well created."""
        client = clients[0]

        dataset_query = asset_factory.create_dataset()
        dataset_key = client.add_dataset(dataset_query)

        data_sample = asset_factory.create_data_sample(datasets=[dataset_key])
        sample_key = client.add_data_sample(data_sample)

        function_query = asset_factory.create_function(FunctionCategory.simple)
        function_key = client.add_function(function_query)

        predict_function_spec = asset_factory.create_function(category=FunctionCategory.predict)
        predict_function_key = client.add_function(predict_function_spec)

        metric = asset_factory.create_function(category=FunctionCategory.metric)
        metric_key = client.add_function(metric)

        cp = asset_factory.create_compute_plan()

        traintask = substra.sdk.schemas.ComputePlanTaskSpec(
            function_key=function_key,
            task_id=str(uuid.uuid4()),
            inputs=FLTaskInputGenerator.task(opener_key=dataset_key, data_sample_keys=[sample_key]),
            outputs=FLTaskOutputGenerator.traintask(),
            worker=client.organization_info().organization_id,
        )

        predicttask = substra.sdk.schemas.ComputePlanTaskSpec(
            function_key=predict_function_key,
            task_id=str(uuid.uuid4()),
            inputs=FLTaskInputGenerator.task(opener_key=dataset_key, data_sample_keys=[sample_key])
            + FLTaskInputGenerator.train_to_predict(traintask.task_id),
            outputs=FLTaskOutputGenerator.predicttask(),
            worker=client.organization_info().organization_id,
        )

        testtasks = [
            substra.sdk.schemas.ComputePlanTaskSpec(
                task_id=str(uuid.uuid4()),
                function_key=metric_key,
                inputs=FLTaskInputGenerator.task(opener_key=dataset_key, data_sample_keys=[sample_key])
                + FLTaskInputGenerator.predict_to_test(predicttask.task_id),
                outputs=FLTaskOutputGenerator.testtask(),
                worker=client.organization_info().organization_id,
            )
        ]
        cp.tasks = [traintask, predicttask] + testtasks

        client.add_compute_plan(cp)

        json_perf_path = client.temp_directory.parent / "live_performances" / cp.key / "performances.json"

        assert json_perf_path.is_file()


class TestsList:
    "Test client.list... functions"

    @pytest.fixture()
    def _init_data_samples(self, asset_factory, docker_clients):
        client = docker_clients[0]
        data_sample_keys = []
        dataset_query = asset_factory.create_dataset()
        dataset_key = client.add_dataset(dataset_query)

        data_sample_1 = asset_factory.create_data_sample(datasets=[dataset_key])
        data_sample_keys.append(client.add_data_sample(data_sample_1))

        data_sample_2 = asset_factory.create_data_sample(datasets=[dataset_key])
        data_sample_keys.append(client.add_data_sample(data_sample_2))

        data_sample_3 = asset_factory.create_data_sample(datasets=[dataset_key])
        data_sample_keys.append(client.add_data_sample(data_sample_3))
        return client, data_sample_keys

    @pytest.mark.parametrize("asset_name", ["dataset", "function"])
    def test_list_assets(self, asset_name, asset_factory, docker_clients):
        client = docker_clients[0]
        query = getattr(asset_factory, f"create_{asset_name}")()
        key = getattr(client, f"add_{asset_name}")(query)

        assets = getattr(client, f"list_{asset_name}")()

        assert assets[0].key == key

    def test_list_datasamples_all(self, _init_data_samples):
        # get all datasamples
        client, _ = _init_data_samples
        data_samples = client.list_data_sample(ascending=False)
        assert len(data_samples) >= 3
        # assert dates in descending order
        assert all(
            data_samples[i].creation_date >= data_samples[i + 1].creation_date for i in range(len(data_samples) - 1)
        )

    def test_list_datasamples_all_order_ascending(self, _init_data_samples):
        # get all datasamples
        client, _ = _init_data_samples
        data_samples = client.list_data_sample(ascending=True)
        assert len(data_samples) >= 3
        # assert dates in ascending order
        assert all(
            data_samples[i].creation_date <= data_samples[i + 1].creation_date for i in range(len(data_samples) - 1)
        )

    def test_list_datasamples_key(self, _init_data_samples):
        # Get sample 1 by key
        client, data_sample_keys = _init_data_samples
        filters = {"key": [data_sample_keys[0]]}
        data_samples = client.list_data_sample(filters=filters)
        assert len(data_samples) == 1
        assert data_samples[0].key == data_sample_keys[0]

    def test_list_datasamples_OR(self, _init_data_samples):
        # Get sample 1 and 2 by key
        client, data_sample_keys = _init_data_samples
        filters = {"key": [data_sample_keys[0], data_sample_keys[1]]}
        data_samples = client.list_data_sample(filters=filters)
        assert len(data_samples) == 2
        assert data_samples[0].key == data_sample_keys[1]
        assert data_samples[1].key == data_sample_keys[0]

    def test_list_datasamples_AND(self, _init_data_samples):
        # Test the AND operator
        # Get sample 1 by key
        client, data_sample_keys = _init_data_samples
        filters = {"key": [data_sample_keys[0]]}
        data_samples = client.list_data_sample(filters=filters)
        assert len(data_samples) == 1
        assert data_samples[0].key == data_sample_keys[0]

        # Get sample 0 by key and wrong owner value (should send empty list)
        client, data_sample_keys = _init_data_samples
        filters = {"key": [data_sample_keys[0]], "owner": ["fake_owner"]}
        data_samples = client.list_data_sample(filters=filters)
        assert len(data_samples) == 0

    def test_list_filter_on_field(self, docker_clients, asset_factory):
        client = docker_clients[0]
        dataset_keys = set()

        n_datasets = 2

        for _ in range(n_datasets):
            dataset_query = asset_factory.create_dataset(name="dataset_dummy_")
            dataset_key = client.add_dataset(dataset_query)
            dataset_keys.add(dataset_key)

        # Get all samples with test_only
        filters = {"name": "dataset_dummy_"}
        datasets = client.list_dataset(filters=filters)
        assert len(datasets) == n_datasets
        assert set([d.key for d in datasets]) == dataset_keys


def test_execute_compute_plan_several_testtasks_per_train(asset_factory, subprocess_clients):
    client = subprocess_clients[0]

    dataset_query = asset_factory.create_dataset()
    dataset_key = client.add_dataset(dataset_query)

    data_sample_1 = asset_factory.create_data_sample(datasets=[dataset_key])
    sample_1_key = client.add_data_sample(data_sample_1)

    function_query = asset_factory.create_function(FunctionCategory.simple)
    function_key = client.add_function(function_query)

    predict_function_spec = asset_factory.create_function(category=FunctionCategory.predict)
    predict_function_key = client.add_function(predict_function_spec)

    metric = asset_factory.create_function(category=FunctionCategory.metric)
    metric_key = client.add_function(metric)

    cp = asset_factory.create_compute_plan()

    traintask = substra.sdk.schemas.ComputePlanTaskSpec(
        function_key=function_key,
        task_id=str(uuid.uuid4()),
        inputs=FLTaskInputGenerator.task(opener_key=dataset_key, data_sample_keys=[sample_1_key]),
        outputs=FLTaskOutputGenerator.traintask(),
        worker=client.organization_info().organization_id,
    )

    predicttask = substra.sdk.schemas.ComputePlanTaskSpec(
        function_key=predict_function_key,
        task_id=str(uuid.uuid4()),
        inputs=FLTaskInputGenerator.task(opener_key=dataset_key, data_sample_keys=[sample_1_key])
        + FLTaskInputGenerator.train_to_predict(traintask.task_id),
        outputs=FLTaskOutputGenerator.predicttask(),
        worker=client.organization_info().organization_id,
    )

    testtasks = [
        substra.sdk.schemas.ComputePlanTaskSpec(
            task_id=str(uuid.uuid4()),
            function_key=metric_key,
            inputs=FLTaskInputGenerator.task(opener_key=dataset_key, data_sample_keys=[sample_1_key])
            + FLTaskInputGenerator.predict_to_test(predicttask.task_id),
            outputs=FLTaskOutputGenerator.testtask(),
            worker=client.organization_info().organization_id,
        )
        for _ in range(2)
    ]
    cp.tasks = [traintask, predicttask] + testtasks

    compute_plan = client.add_compute_plan(cp)
    assert compute_plan.done_count == 4

    tasks = client.list_task(filters={"compute_plan_key": [compute_plan.key]})
    assert len(tasks) == 4


def test_two_composite_to_composite(asset_factory, subprocess_clients):
    client = subprocess_clients[0]

    dataset_query = asset_factory.create_dataset()
    dataset_key = client.add_dataset(dataset_query)

    data_sample_1 = asset_factory.create_data_sample(datasets=[dataset_key])
    sample_1_key = client.add_data_sample(data_sample_1)

    function_query = asset_factory.create_function(FunctionCategory.composite)
    function_key = client.add_function(function_query)

    predict_function_spec = asset_factory.create_function(category=FunctionCategory.predict_composite)
    predict_function_key = client.add_function(predict_function_spec)

    metric = asset_factory.create_function(category=FunctionCategory.metric)
    metric_key = client.add_function(metric)

    cp = asset_factory.create_compute_plan()

    authorized_ids = ["MyOrg1", "MyOrg2"]

    composite_1_key = str(uuid.uuid4())
    composite_1 = substra.sdk.schemas.ComputePlanTaskSpec(
        function_key=function_key,
        task_id=composite_1_key,
        inputs=FLTaskInputGenerator.task(opener_key=dataset_key, data_sample_keys=[sample_1_key]),
        outputs=FLTaskOutputGenerator.composite_traintask(
            shared_authorized_ids=authorized_ids, local_authorized_ids=authorized_ids
        ),
        worker=client.organization_info().organization_id,
    )
    composite_2_key = str(uuid.uuid4())
    composite_2 = substra.sdk.schemas.ComputePlanTaskSpec(
        function_key=function_key,
        task_id=composite_2_key,
        inputs=FLTaskInputGenerator.task(opener_key=dataset_key, data_sample_keys=[sample_1_key]),
        outputs=FLTaskOutputGenerator.composite_traintask(
            shared_authorized_ids=authorized_ids, local_authorized_ids=authorized_ids
        ),
        worker=client.organization_info().organization_id,
    )

    composite_3_key = str(uuid.uuid4())
    composite_3 = substra.sdk.schemas.ComputePlanTaskSpec(
        function_key=function_key,
        task_id=composite_3_key,
        inputs=FLTaskInputGenerator.task(opener_key=dataset_key, data_sample_keys=[sample_1_key])
        + FLTaskInputGenerator.composite_to_composite(composite_1_key, composite_2_key),
        outputs=FLTaskOutputGenerator.composite_traintask(
            shared_authorized_ids=authorized_ids, local_authorized_ids=authorized_ids
        ),
        worker=client.organization_info().organization_id,
    )

    predicttask_key = str(uuid.uuid4())
    predicttask = substra.sdk.schemas.ComputePlanTaskSpec(
        function_key=predict_function_key,
        task_id=predicttask_key,
        inputs=FLTaskInputGenerator.task(opener_key=dataset_key, data_sample_keys=[sample_1_key])
        + FLTaskInputGenerator.composite_to_predict(composite_3_key),
        outputs=FLTaskOutputGenerator.predicttask(authorized_ids),
        worker=client.organization_info().organization_id,
    )

    testtask = substra.sdk.schemas.ComputePlanTaskSpec(
        function_key=metric_key,
        task_id=str(uuid.uuid4()),
        inputs=FLTaskInputGenerator.task(opener_key=dataset_key, data_sample_keys=[sample_1_key])
        + FLTaskInputGenerator.predict_to_test(predicttask_key),
        outputs=FLTaskOutputGenerator.testtask(authorized_ids),
        worker=client.organization_info().organization_id,
    )

    cp.tasks = [composite_1, composite_2, composite_3, predicttask, testtask]

    compute_plan = client.add_compute_plan(cp)
    assert compute_plan.done_count == 5
    assert client.get_performances(compute_plan.key).performance[0] == 32


def test_task_different_owner_dataset(asset_factory, clients):
    client = clients[0]

    # set dataset, metric and function
    dataset_query = asset_factory.create_dataset()
    dataset_key = client.add_dataset(dataset_query)

    data_sample = asset_factory.create_data_sample(datasets=[dataset_key])
    data_sample_key = client.add_data_sample(data_sample)

    function_query = asset_factory.create_function(FunctionCategory.simple)
    function_key = client.add_function(function_query)

    with pytest.raises(substra.sdk.exceptions.InvalidRequest) as err:
        client.add_task(
            substra.sdk.schemas.TaskSpec(
                function_key=function_key,
                inputs=[
                    substra.sdk.schemas.InputRef(identifier=InputIdentifiers.datasamples, asset_key=data_sample_key),
                    substra.sdk.schemas.InputRef(identifier=InputIdentifiers.opener, asset_key=dataset_key),
                ],
                outputs=FLTaskOutputGenerator.traintask(),
                worker="bad_worker",
            )
        )
    assert "The task worker must be the organization that contains the data:" in str(err.value)


class TestMultipleOrgLocalClient:
    def test_local_org_info(self, docker_clients):
        org1info = docker_clients[0].organization_info()
        org2info = docker_clients[1].organization_info()

        assert org1info.organization_id != org2info.organization_id

    def test_list_org(self, docker_clients):
        org1_id = docker_clients[0].organization_info().organization_id
        org2_id = docker_clients[1].organization_info().organization_id

        orgs = docker_clients[0].list_organization()

        org1 = [x for x in orgs if x.id == org1_id]
        assert len(org1) == 1
        assert org1[0].is_current

        org2 = [x for x in orgs if x.id == org2_id]
        assert len(org2) == 1
        assert not org2[0].is_current

    def test_local_org_owner(self, asset_factory, docker_clients):
        client = docker_clients[0]
        dataset_query = asset_factory.create_dataset()
        dataset_key = client.add_dataset(dataset_query)

        dataset = client.get_dataset(dataset_key)
        assert dataset.owner == client.organization_info().organization_id

    def test_local_org_shared_db(self, asset_factory, docker_clients):
        dataset_query = asset_factory.create_dataset()
        dataset_key = docker_clients[0].add_dataset(dataset_query)

        docker_clients[1].get_dataset(dataset_key)



import abc
from typing import List

from substra.sdk.schemas import BackendType


class BaseBackend(abc.ABC):
    @property
    @abc.abstractmethod
    def backend_mode(self) -> BackendType:
        raise NotImplementedError

    @abc.abstractmethod
    def login(self, username, password):
        pass

    @abc.abstractmethod
    def logout(self):
        pass

    @abc.abstractmethod
    def get(self, asset_type, key):
        raise NotImplementedError

    @abc.abstractmethod
    def list(self, asset_type, filters=None, paginated=False):
        raise NotImplementedError

    @abc.abstractmethod
    def add(self, spec, spec_options=None):
        raise NotImplementedError

    @abc.abstractmethod
    def update(self, key, spec, spec_options=None):
        raise NotImplementedError

    @abc.abstractmethod
    def add_compute_plan_tasks(self, spec, spec_options):
        raise NotImplementedError

    @abc.abstractmethod
    def link_dataset_with_data_samples(self, dataset_key, data_sample_keys) -> List[str]:
        raise NotImplementedError

    @abc.abstractmethod
    def download(self, asset_type, url_field_path, key, destination):
        raise NotImplementedError

    @abc.abstractmethod
    def describe(self, asset_type, key):
        raise NotImplementedError

    @abc.abstractmethod
    def cancel_compute_plan(self, key):
        raise NotImplementedError



from enum import Enum

from substra.sdk.schemas import AssetKind
from substra.sdk.schemas import ComputeTaskOutputSpec
from substra.sdk.schemas import FunctionInputSpec
from substra.sdk.schemas import FunctionOutputSpec
from substra.sdk.schemas import InputRef
from substra.sdk.schemas import Permissions

PUBLIC_PERMISSIONS = Permissions(public=True, authorized_ids=[])


class FunctionCategory(str, Enum):
    """Function category"""

    unknown = "FUNCTION_UNKNOWN"
    simple = "FUNCTION_SIMPLE"
    composite = "FUNCTION_COMPOSITE"
    aggregate = "FUNCTION_AGGREGATE"
    metric = "FUNCTION_METRIC"
    predict = "FUNCTION_PREDICT"
    predict_composite = "FUNCTION_PREDICT_COMPOSITE"


class InputIdentifiers(str, Enum):
    local = "local"
    shared = "shared"
    predictions = "predictions"
    performance = "performance"
    opener = "opener"
    datasamples = "datasamples"


class OutputIdentifiers(str, Enum):
    local = "local"
    shared = "shared"
    predictions = "predictions"
    performance = "performance"


class FLFunctionInputs(list, Enum):
    """Substra function inputs by function category based on the InputIdentifiers"""

    FUNCTION_AGGREGATE = [
        FunctionInputSpec(identifier=InputIdentifiers.shared, kind=AssetKind.model.value, optional=False, multiple=True)
    ]
    FUNCTION_SIMPLE = [
        FunctionInputSpec(
            identifier=InputIdentifiers.datasamples,
            kind=AssetKind.data_sample.value,
            optional=False,
            multiple=True,
        ),
        FunctionInputSpec(
            identifier=InputIdentifiers.opener, kind=AssetKind.data_manager.value, optional=False, multiple=False
        ),
        FunctionInputSpec(identifier=InputIdentifiers.shared, kind=AssetKind.model.value, optional=True, multiple=True),
    ]
    FUNCTION_COMPOSITE = [
        FunctionInputSpec(
            identifier=InputIdentifiers.datasamples,
            kind=AssetKind.data_sample.value,
            optional=False,
            multiple=True,
        ),
        FunctionInputSpec(
            identifier=InputIdentifiers.opener, kind=AssetKind.data_manager.value, optional=False, multiple=False
        ),
        FunctionInputSpec(identifier=InputIdentifiers.local, kind=AssetKind.model.value, optional=True, multiple=False),
        FunctionInputSpec(
            identifier=InputIdentifiers.shared, kind=AssetKind.model.value, optional=True, multiple=False
        ),
    ]
    FUNCTION_PREDICT = [
        FunctionInputSpec(
            identifier=InputIdentifiers.datasamples,
            kind=AssetKind.data_sample.value,
            optional=False,
            multiple=True,
        ),
        FunctionInputSpec(
            identifier=InputIdentifiers.opener, kind=AssetKind.data_manager.value, optional=False, multiple=False
        ),
        FunctionInputSpec(
            identifier=InputIdentifiers.shared, kind=AssetKind.model.value, optional=False, multiple=False
        ),
    ]
    FUNCTION_PREDICT_COMPOSITE = [
        FunctionInputSpec(
            identifier=InputIdentifiers.datasamples,
            kind=AssetKind.data_sample.value,
            optional=False,
            multiple=True,
        ),
        FunctionInputSpec(
            identifier=InputIdentifiers.opener, kind=AssetKind.data_manager.value, optional=False, multiple=False
        ),
        FunctionInputSpec(
            identifier=InputIdentifiers.local, kind=AssetKind.model.value, optional=False, multiple=False
        ),
        FunctionInputSpec(
            identifier=InputIdentifiers.shared, kind=AssetKind.model.value, optional=False, multiple=False
        ),
    ]
    FUNCTION_METRIC = [
        FunctionInputSpec(
            identifier=InputIdentifiers.datasamples,
            kind=AssetKind.data_sample.value,
            optional=False,
            multiple=True,
        ),
        FunctionInputSpec(
            identifier=InputIdentifiers.opener, kind=AssetKind.data_manager.value, optional=False, multiple=False
        ),
        FunctionInputSpec(
            identifier=InputIdentifiers.predictions, kind=AssetKind.model.value, optional=False, multiple=False
        ),
    ]


class FLFunctionOutputs(list, Enum):
    """Substra function outputs by function category based on the OutputIdentifiers"""

    FUNCTION_AGGREGATE = [
        FunctionOutputSpec(identifier=OutputIdentifiers.shared, kind=AssetKind.model.value, multiple=False)
    ]
    FUNCTION_SIMPLE = [
        FunctionOutputSpec(identifier=OutputIdentifiers.shared, kind=AssetKind.model.value, multiple=False)
    ]
    FUNCTION_COMPOSITE = [
        FunctionOutputSpec(identifier=OutputIdentifiers.local, kind=AssetKind.model.value, multiple=False),
        FunctionOutputSpec(identifier=OutputIdentifiers.shared, kind=AssetKind.model.value, multiple=False),
    ]
    FUNCTION_PREDICT = [
        FunctionOutputSpec(identifier=OutputIdentifiers.predictions, kind=AssetKind.model.value, multiple=False)
    ]
    FUNCTION_PREDICT_COMPOSITE = [
        FunctionOutputSpec(identifier=OutputIdentifiers.predictions, kind=AssetKind.model.value, multiple=False)
    ]
    FUNCTION_METRIC = [
        FunctionOutputSpec(identifier=OutputIdentifiers.performance, kind=AssetKind.performance.value, multiple=False)
    ]


class FLTaskInputGenerator:
    "Generates task inputs based on Input and OutputIdentifiers"

    @staticmethod
    def opener(opener_key):
        return [InputRef(identifier=InputIdentifiers.opener, asset_key=opener_key)]

    @staticmethod
    def data_samples(data_sample_keys):
        return [
            InputRef(identifier=InputIdentifiers.datasamples, asset_key=data_sample) for data_sample in data_sample_keys
        ]

    @staticmethod
    def task(opener_key, data_sample_keys):
        return FLTaskInputGenerator.opener(opener_key=opener_key) + FLTaskInputGenerator.data_samples(
            data_sample_keys=data_sample_keys
        )

    @staticmethod
    def trains_to_train(model_keys):
        return [
            InputRef(
                identifier=InputIdentifiers.shared,
                parent_task_key=model_key,
                parent_task_output_identifier=OutputIdentifiers.shared,
            )
            for model_key in model_keys
        ]

    @staticmethod
    def trains_to_aggregate(model_keys):
        return [
            InputRef(
                identifier=InputIdentifiers.shared,
                parent_task_key=model_key,
                parent_task_output_identifier=OutputIdentifiers.shared,
            )
            for model_key in model_keys
        ]

    @staticmethod
    def train_to_predict(model_key):
        return [
            InputRef(
                identifier=InputIdentifiers.shared,
                parent_task_key=model_key,
                parent_task_output_identifier=OutputIdentifiers.shared,
            )
        ]

    @staticmethod
    def predict_to_test(model_key):
        return [
            InputRef(
                identifier=InputIdentifiers.predictions,
                parent_task_key=model_key,
                parent_task_output_identifier=OutputIdentifiers.predictions,
            )
        ]

    @staticmethod
    def composite_to_predict(model_key):
        return [
            InputRef(
                identifier=InputIdentifiers.local,
                parent_task_key=model_key,
                parent_task_output_identifier=OutputIdentifiers.local,
            ),
            InputRef(
                identifier=InputIdentifiers.shared,
                parent_task_key=model_key,
                parent_task_output_identifier=OutputIdentifiers.shared,
            ),
        ]

    @staticmethod
    def composite_to_local(model_key):
        return [
            InputRef(
                identifier=InputIdentifiers.local,
                parent_task_key=model_key,
                parent_task_output_identifier=OutputIdentifiers.local,
            )
        ]

    @staticmethod
    def composite_to_composite(model1_key, model2_key=None):
        return [
            InputRef(
                identifier=InputIdentifiers.local,
                parent_task_key=model1_key,
                parent_task_output_identifier=OutputIdentifiers.local,
            ),
            InputRef(
                identifier=InputIdentifiers.shared,
                parent_task_key=model2_key or model1_key,
                parent_task_output_identifier=OutputIdentifiers.shared,
            ),
        ]

    @staticmethod
    def aggregate_to_shared(model_key):
        return [
            InputRef(
                identifier=InputIdentifiers.shared,
                parent_task_key=model_key,
                parent_task_output_identifier=OutputIdentifiers.shared,
            )
        ]

    @staticmethod
    def composites_to_aggregate(model_keys):
        return [
            InputRef(
                identifier=InputIdentifiers.shared,
                parent_task_key=model_key,
                parent_task_output_identifier=OutputIdentifiers.shared,
            )
            for model_key in model_keys
        ]

    @staticmethod
    def aggregate_to_predict(model_key):
        return [
            InputRef(
                identifier=InputIdentifiers.shared,
                parent_task_key=model_key,
                parent_task_output_identifier=OutputIdentifiers.shared,
            )
        ]

    @staticmethod
    def local_to_aggregate(model_key):
        return [
            InputRef(
                identifier=InputIdentifiers.shared,
                parent_task_key=model_key,
                parent_task_output_identifier=OutputIdentifiers.local,
            )
        ]

    @staticmethod
    def shared_to_aggregate(model_key):
        return [
            InputRef(
                identifier=InputIdentifiers.shared,
                parent_task_key=model_key,
                parent_task_output_identifier=OutputIdentifiers.shared,
            )
        ]


def _permission_from_ids(authorized_ids):
    if authorized_ids is None:
        return PUBLIC_PERMISSIONS

    return Permissions(public=False, authorized_ids=authorized_ids)


class FLTaskOutputGenerator:
    "Generates task outputs based on Input and OutputIdentifiers"

    @staticmethod
    def traintask(authorized_ids=None):
        return {OutputIdentifiers.shared: ComputeTaskOutputSpec(permissions=_permission_from_ids(authorized_ids))}

    @staticmethod
    def aggregatetask(authorized_ids=None):
        return {OutputIdentifiers.shared: ComputeTaskOutputSpec(permissions=_permission_from_ids(authorized_ids))}

    @staticmethod
    def predicttask(authorized_ids=None):
        return {OutputIdentifiers.predictions: ComputeTaskOutputSpec(permissions=_permission_from_ids(authorized_ids))}

    @staticmethod
    def testtask(authorized_ids=None):
        return {
            OutputIdentifiers.performance: ComputeTaskOutputSpec(
                permissions=Permissions(public=True, authorized_ids=[])
            )
        }

    @staticmethod
    def composite_traintask(shared_authorized_ids=None, local_authorized_ids=None):
        return {
            OutputIdentifiers.shared: ComputeTaskOutputSpec(permissions=_permission_from_ids(shared_authorized_ids)),
            OutputIdentifiers.local: ComputeTaskOutputSpec(permissions=_permission_from_ids(local_authorized_ids)),
        }



"""Packaging settings."""
import os
from codecs import open

from setuptools import find_packages
from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))


with open(os.path.join(here, "README.md"), "r", "utf-8") as fp:
    readme = fp.read()


about = {}
with open(os.path.join(here, "substra", "__version__.py"), "r", "utf-8") as fp:
    exec(fp.read(), about)


setup(
    name="substra",
    version=about["__version__"],
    description="Low-level Python library for interacting with a Substra network",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://docs.substra.org",
    author="Owkin, Inc.",
    license="Apache 2.0",
    classifiers=[
        "Intended Audience :: Developers",
        "Topic :: Utilities",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    keywords=["cli", "substra"],
    packages=find_packages(exclude=["tests*"]),
    include_package_data=True,
    install_requires=[
        "requests",
        "urllib3<2",
        "docker",
        "pyyaml",
        "pydantic>=2.3.0,<3.0.0",
        "tqdm",
        "python-slugify",
    ],
    python_requires=">=3.8",
    extras_require={
        "dev": [
            "pandas",
            "pytest",
            "pytest-cov",
            "pytest-mock",
            "substratools>=0.20.0",
            "black",
            "flake8",
            "isort",
            "docstring_parser",
        ],
    },
    zip_safe=False,
)



import json
import logging
import math
from copy import deepcopy
from typing import Dict
from typing import List
from typing import Union

from substra.sdk import compute_plan
from substra.sdk import exceptions
from substra.sdk import models
from substra.sdk import schemas
from substra.sdk.backends import base
from substra.sdk.backends.remote import rest_client

logger = logging.getLogger(__name__)

DEFAULT_RETRY_TIMEOUT = 5 * 60
AUTO_BATCHING = "auto_batching"
BATCH_SIZE = "batch_size"
DOWNLOAD_CHUNK_SIZE = 1024


def _find_asset_field(data, field):
    """Find data value where location is defined as `field.subfield...`."""
    for f in field.split("."):
        data = getattr(data, f)
    return data


class Remote(base.BaseBackend):
    def __init__(self, url, insecure, token, retry_timeout, backend_type):
        self._client = rest_client.Client(url, insecure, token)
        self._retry_timeout = retry_timeout or DEFAULT_RETRY_TIMEOUT
        assert backend_type == self.backend_mode

    @property
    def backend_mode(self) -> schemas.BackendType:
        """Get the backend mode: remote"""
        return schemas.BackendType.REMOTE

    def login(self, username, password):
        return self._client.login(username, password)

    def logout(self):
        return self._client.logout()

    def get(self, asset_type, key):
        """Get an asset by key."""
        asset = self._client.get(asset_type.to_server(), key)
        return models.SCHEMA_TO_MODEL[asset_type](**asset)

    def get_task_output_asset(self, compute_task_key: str, identifier: str) -> models.OutputAsset:
        outputs = self._client.list(
            schemas.Type.Task.to_server(),
            path=compute_task_key + "/output_assets",
            filters={"identifier": [identifier]},
        )

        if len(outputs) == 0:
            raise exceptions.TaskAssetNotFoundError(compute_task_key=compute_task_key, identifier=identifier)
        elif len(outputs) > 1:
            raise exceptions.TaskAssetMultipleFoundError(compute_task_key=compute_task_key, identifier=identifier)
        return models.OutputAsset(**outputs[0])

    def list_task_output_assets(self, compute_task_key: str) -> List[models.OutputAsset]:
        outputs = self._client.list(
            schemas.Type.Task.to_server(),
            path=compute_task_key + "/output_assets",
        )
        return [models.OutputAsset(**output) for output in outputs]

    def list_task_input_assets(self, compute_task_key: str) -> List[models.InputAsset]:
        inputs = self._client.list(
            schemas.Type.Task.to_server(),
            path=compute_task_key + "/input_assets",
        )
        return [models.InputAsset(**i) for i in inputs]

    def get_performances(self, key: str) -> models.Performances:
        """Get an compute plan performance by key."""

        compute_plan = self.get(schemas.Type.ComputePlan, key)
        response = self._client.list(schemas.Type.ComputePlan.to_server(), path=key + "/perf", paginated=False)
        results = response["results"]

        performances = models.Performances()

        for test_task in results:
            performances.compute_plan_key.append(compute_plan.key)
            performances.compute_plan_tag.append(compute_plan.tag)
            performances.compute_plan_status.append(compute_plan.status)
            performances.compute_plan_start_date.append(compute_plan.start_date)
            performances.compute_plan_end_date.append(compute_plan.end_date)
            performances.compute_plan_metadata.append(compute_plan.metadata)

            performances.worker.append(test_task["compute_task"]["worker"])
            performances.task_key.append(test_task["compute_task"]["key"])
            performances.task_rank.append(test_task["compute_task"]["rank"])
            try:
                round_idx = int(test_task["compute_task"]["round_idx"])
            except TypeError:
                round_idx = None
            performances.round_idx.append(round_idx)
            performances.identifier.append(test_task["identifier"])
            performances.performance.append(test_task["perf"])

        return performances

    def list(
        self,
        asset_type: schemas.Type,
        filters: Dict[str, Union[List[str], dict]] = None,
        order_by: str = None,
        ascending: bool = False,
        paginated: bool = True,
    ) -> List[models._Model]:
        """List assets of asset_type with filters.

        Args:
            cf [rest_client](substra.sdk.backends.rest_client.list)

        Returns:
            List[models._Model] : a List of assets
        """
        assets = self._client.list(
            asset_type=asset_type.to_server(),
            filters=filters,
            order_by=order_by,
            ascending=ascending,
            paginated=paginated,
        )
        return [models.SCHEMA_TO_MODEL[asset_type](**asset) for asset in assets]

    def _add(self, asset, data, files=None):
        data = deepcopy(data)  # make a deep copy for avoiding modification by reference
        if files:
            kwargs = {
                "data": {
                    "json": json.dumps(data),
                },
                "files": files,
            }

        else:
            kwargs = {
                "json": data,
            }

        return self._client.add(
            asset.to_server(),
            retry_timeout=self._retry_timeout,
            **kwargs,
        )

    def _add_data_samples(self, spec, spec_options):
        """Add data sample(s)."""
        # data sample(s) creation must be handled separately as the get request is not
        # available on this ressource.
        try:
            with spec.build_request_kwargs(**spec_options) as (data, files):
                data_samples = self._add(schemas.Type.DataSample, data, files)
        except exceptions.AlreadyExists as e:
            if spec.is_many():
                # We don't know which of the keys already exists
                raise

            key = e.key[0]
            logger.warning(f"data_sample already exists: key='{key}'")
            data_samples = [{"key": key}]

        # there is currently a single route in the backend to add a single or many
        # datasamples, this route always returned a list of created data sample keys
        return [data_sample["key"] for data_sample in data_samples] if spec.is_many() else data_samples[0]["key"]

    def add(self, spec, spec_options=None):
        """Add an asset."""
        spec_options = spec_options or {}
        asset_type = spec.__class__.type_
        # Remove auto_batching and batch_size from spec_options
        auto_batching = spec_options.pop(AUTO_BATCHING, False)
        batch_size = spec_options.pop(BATCH_SIZE, None)

        if asset_type == schemas.Type.DataSample:
            # data sample corner case
            return self._add_data_samples(
                spec,
                spec_options=spec_options,
            )
        elif asset_type == schemas.Type.ComputePlan:
            cp = self._add_compute_plan(spec, spec_options)
            self._add_tasks_from_computeplan(spec, spec_options, auto_batching, batch_size)
            return cp
        elif asset_type == schemas.Type.Task:
            cp_spec = schemas.ComputePlanSpec(name=f"{spec.key}")
            self._add_compute_plan(cp_spec, spec_options)
            spec.compute_plan_key = cp_spec.key
            return self._add_tasks([spec], spec_options)[0]["key"]

        with spec.build_request_kwargs(**spec_options) as (data, files):
            response = self._add(asset_type, data, files=files)

        return response["key"]

    def _add_compute_plan(self, spec, spec_options):
        """Register compute plan info (without tasks)."""
        cp_spec = spec.model_copy()
        del cp_spec.tasks

        with spec.build_request_kwargs(**spec_options) as (data, _):
            response = self._add(schemas.Type.ComputePlan, data)
        return models.ComputePlan(**response)

    def _add_tasks_from_computeplan(self, spec, spec_options, auto_batching, batch_size):
        """Register batch(es) of tasks."""
        tasks = compute_plan.get_tasks(spec)
        if auto_batching:
            if not batch_size:
                raise ValueError(
                    "Batch size must be defined to create a compute plan \
                    with the auto-batching feature."
                )
            # Split tasks by batch
            batches = []
            for i in range(math.ceil(len(tasks) / batch_size)):
                start = i * batch_size
                end = min(len(tasks), (i + 1) * batch_size)
                batches.append(tasks[start:end])
        else:
            batches = [tasks]

        for batch in batches:
            try:
                self._add_tasks(batch, spec_options)
            except exceptions.AlreadyExists:
                logger.warning(
                    "Skipping already submitted tasks, probably because of a timeout error. "
                    "Check that the compute plan has the right number of tasks once the submission is complete."
                )
                continue

    def _add_tasks(self, batch, spec_options):
        batch_data = []
        for task_spec in batch:
            with task_spec.build_request_kwargs(**spec_options) as (task_data, _):
                batch_data.append(task_data)
        return self._client.request(
            "post",
            "task",
            path="bulk_create/",
            json={"tasks": batch_data},
        )

    def _update(self, asset, key, data, files=None):
        data = deepcopy(data)  # make a deep copy for avoiding modification by reference
        if files:
            kwargs = {
                "data": {
                    "json": json.dumps(data),
                },
                "files": files,
            }

        else:
            kwargs = {
                "json": data,
            }

        return self._client.update(
            asset.to_server(),
            key,
            retry_timeout=self._retry_timeout,
            **kwargs,
        )

    def update(self, key, spec, spec_options=None):
        spec_options = spec_options or {}
        asset_type = spec.__class__.type_
        with spec.build_request_kwargs(**spec_options) as (data, files):
            return self._update(asset_type, key, data, files=files)

    def add_compute_plan_tasks(self, spec, spec_options):
        # Remove auto_batching and batch_size from spec_options
        auto_batching = spec_options.pop(AUTO_BATCHING, False)
        batch_size = spec_options.pop(BATCH_SIZE, None)
        self._add_tasks_from_computeplan(spec, spec_options, auto_batching, batch_size)

        return self.get(
            asset_type=schemas.Type.ComputePlan,
            key=spec.key,
        )

    def link_dataset_with_data_samples(self, dataset_key, data_sample_keys) -> List[str]:
        """Returns the list of the data sample keys"""
        data = {
            "data_manager_keys": [dataset_key],
            "data_sample_keys": data_sample_keys,
        }
        self._client.request(
            "post",
            schemas.Type.DataSample.to_server(),
            path="bulk_update/",
            data=data,
        )
        return data_sample_keys

    def _download(self, url: str, destination_file: str) -> str:
        response = self._client.get_data(url, stream=True)
        with open(destination_file, "wb") as f:
            for chunk in response.iter_content(DOWNLOAD_CHUNK_SIZE):
                f.write(chunk)

        return destination_file

    def download(self, asset_type: schemas.Type, url_field_path: str, key: str, destination: str) -> str:
        data = self.get(asset_type, key)
        url = _find_asset_field(data, url_field_path)
        return self._download(url, destination)

    def download_model(self, key: str, destination_file: str) -> str:
        url = f"{self._client.base_url}/model/{key}/file/"
        return self._download(url, destination_file)

    def download_logs(self, task_key: str, destination_file: str = None) -> str:
        """Download the logs of a failed task. If destination_file is set, return the full
        destination path, otherwise, return the logs as a str.
        """
        url = f"{self._client.base_url}/logs/{task_key}/file/"

        if destination_file:
            return self._download(url, destination_file)

        return self._client.get_data(url).text

    def describe(self, asset_type, key):
        data = self.get(asset_type, key)
        url = data.description.storage_address
        r = self._client.get_data(url)
        return r.text

    def organization_info(self):
        response = self._client.get_data(f"{self._client.base_url}/info/")
        return models.OrganizationInfo(**response.json())

    def cancel_compute_plan(self, key: str) -> None:
        self._client.request(
            "post",
            schemas.Type.ComputePlan.to_server(),
            path=f"{key}/cancel",
        )



import uuid
from typing import List

from pydantic import Field

from substra.sdk import models
from substra.sdk import schemas


class _TaskAssetLocal(schemas._PydanticConfig):
    key: str = Field(default_factory=uuid.uuid4)
    compute_task_key: str

    @staticmethod
    def allowed_filters() -> List[str]:
        return super().allowed_filters() + ["compute_task_key"]


class OutputAssetLocal(models.OutputAsset, _TaskAssetLocal):
    pass


class InputAssetLocal(models.InputAsset, _TaskAssetLocal):
    pass
